"""
Main entry point to either :ref:`install <install-cli>`, :ref:`uninstall <uninstall-cli>`,
:ref:`update <update-cli>`, :ref:`list <list-cli>` or :ref:`create <create-cli>`
MSL packages using the command-line interface (CLI).
"""
import os
import re
import sys
import logging
import subprocess
from pkg_resources import parse_version

from . import utils, __version__, _PKG_NAME

PARSER = None

_pip_options_regexg = re.compile(r'\s+(-[a-zA-Z])?,?\s+(--[-\w]+)\s+(<.*>)?\s')


def configure_parser():
    """:class:`~msl.package_manager.cli_argparse.ArgumentParser`: Returns the argument parser."""

    # pretty much mimics the ArgumentParser structure used by conda

    global PARSER
    if PARSER is not None:
        return PARSER

    from .cli_argparse import ArgumentParser
    from .cli_install import add_parser_install
    from .cli_uninstall import add_parser_uninstall
    from .cli_update import add_parser_update
    from .cli_list import add_parser_list
    from .cli_create import add_parser_create
    from .cli_authorise import add_parser_authorise

    PARSER = ArgumentParser(description='Install, uninstall, update, list or create MSL packages.')

    PARSER.add_argument(
        '-V', '--version',
        action='version',
        version='{} {}'.format(_PKG_NAME, __version__),
        help='Show the MSL Package Manager version number and exit.'
    )

    command_parser = PARSER.add_subparsers(
        metavar='command',
        dest='cmd',
    )
    # https://bugs.python.org/issue9253
    # https://stackoverflow.com/a/18283730/1599393
    command_parser.required = True

    add_parser_install(command_parser)
    add_parser_uninstall(command_parser)
    add_parser_uninstall(command_parser, name='remove')
    add_parser_update(command_parser)
    add_parser_update(command_parser, name='upgrade')
    add_parser_list(command_parser)
    add_parser_create(command_parser)
    add_parser_authorise(command_parser)
    add_parser_authorise(command_parser, name='authorize')

    return PARSER


def parse_args(args):
    """Parse arguments.

    Parameters
    ----------
    args : :class:`list` of :class:`str`
        The arguments to parse.

    Returns
    -------
    An :class:`argparse.Namespace` or :data:`None` if there was an error.
    """
    def pip_options_valid(command):
        # returns either True or False
        valid_options = {}
        out = subprocess.check_output([sys.executable, '-m', 'pip', 'help', command], stderr=subprocess.PIPE)
        for short, long, value in _pip_options_regexg.findall(out.decode()):
            if short:
                valid_options[short] = value
            valid_options[long] = value

        pip_options_copy = pip_options[:]
        for i, option in enumerate(pip_options_copy):
            if not option.startswith('-'):
                # make sure the previous item in pip_options accepts a value
                # if it doesn't then append the option to parsed_args.names
                # because the name of an msl package got mistakenly added to pip_options
                if not valid_options.get(pip_options_copy[i-1]):
                    name = pip_options.pop(pip_options.index(option))
                    parsed_args.names.append(name)

            elif option not in valid_options:
                utils.log.error('No such option for "pip {}": {}'.format(command, option))
                return False

            # if the pip option requires a value then make sure the value did
            # not end up in the parsed_args.names list and the value comes
            # immediately after the appropriate item in the pip_options list
            elif valid_options[option]:
                arg_index = args.index(option) + 1
                pip_index = pip_options_copy.index(option) + 1
                if pip_index >= len(pip_options_copy) or args[arg_index] != pip_options_copy[pip_index]:
                    names_index = parsed_args.names.index(args[arg_index])
                    name = parsed_args.names.pop(names_index)
                    pip_options.insert(pip_index, name)

        return True

    parser = configure_parser()
    # Want to be able to support all options that "pip (un)install"
    # supports without depending on the version of pip that is installed.
    # There are possible 3 ways to do this:
    #   1) create a new argparse with nargs=*
    #   2) create a new argparse with nargs=REMAINDER
    #   3) use parse_known_args() instead of parse_args()
    # Option 3 was the preferred way because nargs=* and nargs=REMAINDER
    # each had their limitations.
    parsed_args, pip_options = parser.parse_known_args(args=args)

    if parsed_args.quiet == 0:
        utils.set_log_level(logging.DEBUG)
    elif parsed_args.quiet == 1:
        utils.set_log_level(logging.INFO)
    elif parsed_args.quiet == 2:
        utils.set_log_level(logging.WARNING)
    elif parsed_args.quiet == 3:
        utils.set_log_level(logging.ERROR)
    elif parsed_args.quiet == 4:
        utils.set_log_level(logging.CRITICAL + 1)

    if pip_options:
        if parsed_args.cmd in ['install', 'update', 'upgrade']:
            if not pip_options_valid('install'):
                return
        elif parsed_args.cmd in ['uninstall', 'remove']:
            if not pip_options_valid('uninstall'):
                return
        else:
            utils.log.warning('The following options are ignored: ' + ', '.join(p for p in pip_options))

    parsed_args.pip_options = pip_options
    return parsed_args


def _main(*args):
    # parse the input
    if not args:
        args = sys.argv[1:]
        if not args:
            args = ['--help']

    args = parse_args(args)
    if not args:
        return

    # when the msl-package-manager gets updated a msl.exe.old file gets created (on Windows)
    old = sys.exec_prefix + '/Scripts/msl.exe.old'
    if os.path.isfile(old):
        os.remove(old)

    # execute the command
    ret = args.func(args, PARSER)
    if ret == 'updating_msl_package_manager':
        return

    if args.disable_mslpm_version_check:
        return

    # check if there is an update for the MSL Package Manager
    if utils.log.level < logging.WARNING:  # do not log DEBUG messages when checking for the update
        utils.set_log_level(logging.WARNING)

    pkgs = utils.pypi()
    if not pkgs:
        return

    latest = pkgs[_PKG_NAME]['version']
    if parse_version(latest) > parse_version(__version__):
        utils.log.warning('You are using {0} version {1}, however, version {2} is available.\n'
                          'You should consider updating via the \'msl update package-manager\''
                          ' command.'.format(_PKG_NAME, __version__, latest))


def main(*args):
    """
    Main entry point to either :ref:`install <install-cli>`, :ref:`uninstall <uninstall-cli>`,
    :ref:`update <update-cli>`, :ref:`list <list-cli>` or :ref:`create <create-cli>`
    MSL packages using the CLI.
    """
    sys.exit(_main(*args))


if __name__ == '__main__':
    _main()



"""
Command line interface for the :ref:`install <install-cli>` command.
"""
from .install import install
from .cli_argparse import (
    add_argument_all,
    add_argument_branch,
    add_argument_package_names,
    add_argument_tag,
    add_argument_quiet,
    add_argument_update_cache,
    add_argument_yes,
    add_argument_disable_mslpm_version_check,
    add_argument_commit,
)

HELP = 'Install MSL packages.'

DESCRIPTION = HELP + """

MSL packages are retrieved from GitHub repositories or from PyPI.
"""

EXAMPLE = """All other optional arguments are passed to "pip install".

Examples:
    msl install equipment
    msl install loadlib --tag v0.3.0
    msl install io network --retries 10
"""


def add_parser_install(parser):
    """Add the :ref:`install <install-cli>` command to the parser."""
    p = parser.add_parser(
        'install',
        help=HELP,
        description=DESCRIPTION,
        epilog=EXAMPLE,
    )
    add_argument_package_names(p)
    add_argument_all(p)
    add_argument_yes(p)
    add_argument_branch(p)
    add_argument_commit(p)
    add_argument_tag(p)
    add_argument_quiet(p)
    add_argument_update_cache(p)
    add_argument_disable_mslpm_version_check(p)
    p.set_defaults(func=execute)


def execute(args, parser):
    """Executes the :ref:`install <install-cli>` command."""
    if parser.contains_package_names():
        install(
            *args.names,
            yes=args.yes,
            branch=args.branch,
            commit=args.commit,
            tag=args.tag,
            update_cache=args.update_cache,
            pip_options=args.pip_options
        )



"""
Command line interface for the :ref:`list <list-cli>` command.
"""
from .utils import info
from .cli_argparse import (
    add_argument_update_cache,
    add_argument_quiet,
    add_argument_disable_mslpm_version_check,
)

HELP = 'Show the information about MSL packages.'

DESCRIPTION = HELP + """

The information can be either for the installed packages,  
packages that are available as GitHub repositories, or
packages available on PyPI.
"""

EXAMPLE = """
Examples:
    msl list
    msl list --github --json
    msl list --pypi
"""


def add_parser_list(parser):
    """Add the :ref:`list <list-cli>` command to the parser."""
    p = parser.add_parser(
        'list',
        help=HELP,
        description=DESCRIPTION,
        epilog=EXAMPLE,
    )
    p.add_argument(
        '-p', '--pypi',
        action='store_true',
        default=False,
        help='Show the information about the PyPI packages.'
    )
    p.add_argument(
        '-g', '--github',
        action='store_true',
        default=False,
        help='Show the information about the GitHub repositories.'
    )
    p.add_argument(
        '-j', '--json',
        action='store_true',
        default=False,
        help='Show the information in JSON format.\n'
             'For the GitHub repositories this includes additional\n'
             'information about the branches and tags.'
    )
    add_argument_quiet(p)
    add_argument_update_cache(p)
    add_argument_disable_mslpm_version_check(p)
    p.set_defaults(func=execute)


def execute(args, parser):
    """Executes the :ref:`list <list-cli>` command."""
    info(args.github, args.pypi, args.update_cache, args.json)



"""
Uninstall MSL packages.
"""
import os
import sys
import subprocess
import pkg_resources

from . import utils


def uninstall(*names, **kwargs):
    """Uninstall MSL packages.

    .. versionchanged:: 2.4.0
        Added the `pip_options` keyword argument.

    Parameters
    ----------
    *names
        The name(s) of the MSL package(s) to uninstall. If not specified then
        uninstall all MSL packages (except for the **MSL Package Manager** --
        in which case use ``pip uninstall msl-package-manager``). The
        ``msl-`` prefix can be omitted (e.g., ``'loadlib'`` is equivalent to
        ``'msl-loadlib'``). Also accepts shell-style wildcards (e.g., ``'pr-*'``).
    **kwargs
        * yes -- :class:`bool`
            If :data:`True` then don't ask for confirmation before uninstalling.
            The default is :data:`False` (ask before uninstalling).
        * pip_options -- :class:`list` of :class:`str`
            Optional arguments to pass to the ``pip uninstall`` command,
            e.g., ``['--no-python-version-warning']``

    """
    # TODO Python 2.7 does not support named arguments after using *args
    #  we can define yes=False, pip_options=None in the function signature
    #  when we choose to drop support for Python 2.7
    utils._check_kwargs(kwargs, {'yes', 'pip_options'})

    yes = kwargs.get('yes', False)
    pip_options = kwargs.get('pip_options', [])

    packages = utils._create_uninstall_list(names)
    if not packages:
        utils.log.info('No MSL packages to uninstall')
        return

    # use the word REMOVE since it visibly looks different than UNINSTALL and INSTALL do
    utils._log_install_uninstall_message(packages, 'REMOVED')
    if not (yes or utils._ask_proceed()):
        return

    # After a MSL package gets uninstalled the "msl" namespace gets destroyed.
    # This is a known issue:
    #   https://github.com/pypa/sample-namespace-packages/issues/5
    #   https://github.com/pypa/python-packaging-user-guide/issues/314
    # There are few ways to bypass this issue
    #   1. force all MSL packages to require that the MSL Package Manager is installed
    #      and only the MSL Package Manager contains the namespace __init__.py file
    #   2. modify the setup.py of each namespace package to have "packages=" include
    #      only the child package.
    #   3. what is done below... assume that MSL packages are uninstalled using the
    #      MSL Package Manager and then re-create the __init__.py files after a
    #      package is uninstalled.

    def check_if_namespace_package(package_name):
        for dist in pkg_resources.working_set:
            if dist.project_name in package_name:
                split = package_name.split('-')
                if len(split) != 2:
                    break

                examples_init_file = os.path.join(dist.module_path, split[0], 'examples', '__init__.py')
                if not os.path.isfile(examples_init_file):
                    break

                with open(examples_init_file, 'rt') as fp:
                    examples_init = fp.readlines()

                init_file = os.path.join(dist.module_path, split[0], '__init__.py')
                if not os.path.isfile(init_file):
                    break

                with open(init_file, 'rt') as fp:
                    init = fp.readlines()

                for line in init:
                    if '__path__' in line and 'pkgutil' in line:
                        return True, os.path.dirname(init_file), init, examples_init

        return False, None, None, None

    utils.log.info('')

    exe = [sys.executable, '-m', 'pip', 'uninstall']

    if '--quiet' not in pip_options or '-q' not in pip_options:
        pip_options.extend(['--quiet'] * utils._pip_quiet)
    if '--disable-pip-version-check' not in pip_options:
        pip_options.append('--disable-pip-version-check')
    if '--yes' not in pip_options or '-y' not in pip_options:
        pip_options.append('--yes')

    for pkg in packages:
        is_namespace, path, init, examples_init = check_if_namespace_package(pkg)
        subprocess.call(exe + pip_options + [pkg])
        if is_namespace and os.path.isdir(path):
            with open(os.path.join(path, '__init__.py'), 'wt') as fp:
                fp.writelines(init)
            with open(os.path.join(path, 'examples', '__init__.py'), 'wt') as fp:
                fp.writelines(examples_init)



"""
Update MSL packages.
"""
import os
import re
import sys
import subprocess
from pkg_resources import parse_version

from colorama import Fore

from . import utils, _PKG_NAME


def update(*names, **kwargs):
    """Update MSL packages.

    MSL packages can be updated from PyPI packages_ (only if a release has been
    uploaded to PyPI) or from GitHub repositories_.

    .. note::
       If the MSL packages_ are available on PyPI then PyPI is used as the default
       URI_ to update the package. If you want to force the update to occur
       from the ``main`` branch of the GitHub `repository <https://github.com/MSLNZ>`_
       then set ``branch='main'``. If the package is not available on PyPI
       then the ``main`` branch is used as the default update URI_.

    .. _repositories: https://github.com/MSLNZ
    .. _packages: https://pypi.org/search/?q=%22Measurement+Standards+Laboratory+of+New+Zealand%22
    .. _URI: https://en.wikipedia.org/wiki/Uniform_Resource_Identifier

    .. versionchanged:: 2.4.0
        Added the `pip_options` keyword argument.

    .. versionchanged:: 2.5.0
        Added the `include_non_msl` and `commit` keyword arguments. The default
        name of a repository branch changed to ``main``.

    Parameters
    ----------
    *names
        The name(s) of the MSL package(s) to update. If not specified then
        update all MSL packages. The ``msl-`` prefix can be omitted (e.g.,
        ``'loadlib'`` is equivalent to ``'msl-loadlib'``). Also accepts
        shell-style wildcards (e.g., ``'pr-*'``).
    **kwargs
        * branch -- :class:`str`
            The name of a git branch to use to update the package(s) to.
        * commit -- :class:`str`
            The hash value of a git commit to use to update a package.
        * tag -- :class:`str`
            The name of a git tag to use to update a package.
        * update_cache -- :class:`bool`
            The information about the MSL packages_ that are available on PyPI and about
            the repositories_ that are available on GitHub are cached to use for subsequent
            calls to this function. After 24 hours the cache is automatically updated. Set
            `update_cache` to be :data:`True` to force the cache to be updated when you call
            this function. Default is :data:`False`.
        * yes -- :class:`bool`
            If :data:`True` then don't ask for confirmation before updating.
            The default is :data:`False` (ask before updating).
        * pip_options -- :class:`list` of :class:`str`
            Optional arguments to pass to the ``pip install --upgrade`` command,
            e.g., ``['--upgrade-strategy', 'eager']``
        * include_non_msl -- :class:`bool`
            If :data:`True` then also update all non-MSL packages.
            The default is :data:`False` (only update the specified
            MSL packages). Warning, enable this option with caution.

        .. important::
           If you specify a `branch`, `commit` or `tag` then the update will be forced.
    """
    # TODO Python 2.7 does not support named arguments after using *args
    #  we can define yes=False, branch=None, ...
    #  in the function signature when we choose to drop support for Python 2.7
    utils._check_kwargs(kwargs, {'yes', 'branch', 'commit', 'tag',
                                 'update_cache', 'pip_options', 'include_non_msl', 'all_msl'})

    yes = kwargs.get('yes', False)
    branch = kwargs.get('branch', None)
    commit = kwargs.get('commit', None)
    tag = kwargs.get('tag', None)
    update_cache = kwargs.get('update_cache', False)
    pip_options = kwargs.get('pip_options', [])
    include_non_msl = kwargs.get('include_non_msl', False)
    # do not include 'all_msl' in docstring, it is only used internally by the CLI
    all_msl = kwargs.get('all_msl', False)

    if commit and not utils.has_git:
        utils.log.error('Cannot update from a commit because git is not installed')
        return

    github_suffix = utils._get_github_url_suffix(branch=branch, commit=commit, tag=tag)
    if github_suffix is None:
        return

    # keep the order of the log messages consistent: pypi -> github -> local
    pkgs_pypi = utils.pypi(update_cache=update_cache)
    pkgs_github = utils.github(update_cache=update_cache)
    pkgs_installed = utils.installed()
    pkgs_non_msl = utils.outdated_pypi_packages(pkgs_installed) if include_non_msl else {}
    if not pkgs_github and not pkgs_pypi and not pkgs_non_msl:
        return

    if not names or all_msl:
        # update all installed MSL packages only if not updating non-MSL packages
        packages = pkgs_installed if (all_msl or not include_non_msl) else {}
    else:
        packages = utils._check_wildcards_and_prefix(names, pkgs_installed)

    w_non_msl = [0, 0]
    if pkgs_non_msl:
        for name, values in pkgs_non_msl.items():
            w_non_msl = [
                max(w_non_msl[0], len(name)),
                max(w_non_msl[1], len(values['installed_version']))
            ]

    w = [0, 0]
    msl_pkgs_to_update = dict()
    for name, values in packages.items():

        err_msg = 'Cannot update {!r} -- '.format(name)

        if name not in pkgs_installed:
            utils.log.error(err_msg + 'The package is not installed')
            continue

        installed_version = pkgs_installed[name]['version']

        # use PyPI to update the package (only if the package is available on PyPI)
        using_pypi = name in pkgs_pypi and not (tag or branch or commit)
        repo_name = pkgs_installed[name]['repo_name']

        # an MSL package could have been installed in "editable" mode, i.e., pip install -e .
        # and therefore it might only exist locally until it is pushed to the repository
        repo = pkgs_github.get(repo_name)
        no_repo_err_msg = err_msg + 'The {!r} repository does not exist'.format(repo_name)

        extras_require = values['extras_require'] if values.get('extras_require') is not None else ''

        if commit is not None:
            if not repo:
                utils.log.error(no_repo_err_msg)
                continue
            # just assume that the commit value is okay
            msl_pkgs_to_update[name] = {
                'installed_version': installed_version,
                'using_pypi': False,
                'extras_require': extras_require,
                'version': '[commit:{}]'.format(commit[:7]),
                'repo_name': repo_name,
            }
        elif tag is not None:
            if not repo:
                utils.log.error(no_repo_err_msg)
                continue
            if tag in repo['tags']:
                msl_pkgs_to_update[name] = {
                    'installed_version': installed_version,
                    'using_pypi': False,
                    'extras_require': extras_require,
                    'version': '[tag:{}]'.format(tag),
                    'repo_name': repo_name,
                }
            else:
                utils.log.error(err_msg + 'The {!r} tag does not exist'.format(tag))
                continue
        elif branch is not None:
            if not repo:
                utils.log.error(no_repo_err_msg)
                continue
            if branch in repo['branches']:
                msl_pkgs_to_update[name] = {
                    'installed_version': installed_version,
                    'using_pypi': False,
                    'extras_require': extras_require,
                    'version': '[branch:{}]'.format(branch),
                    'repo_name': repo_name,
                }
            else:
                utils.log.error(err_msg + 'The {!r} branch does not exist'.format(branch))
                continue
        else:
            if using_pypi:
                version = pkgs_pypi[name]['version']
            else:
                if not repo:
                    utils.log.error(no_repo_err_msg)
                    continue
                version = repo['version']

            if not version:
                # a version number must exist on PyPI, so if this occurs it must be for a github repo
                utils.log.error(
                    err_msg + 'The GitHub repository does not contain a release. Specify a branch, commit or tag'
                )
                continue
            elif values.get('version_requested'):
                # this elif must come before the parse_version check
                msl_pkgs_to_update[name] = {
                    'installed_version': installed_version,
                    'using_pypi': using_pypi,
                    'extras_require': extras_require,
                    'version': values['version_requested'],
                    'repo_name': repo_name,
                }
            elif parse_version(version) > parse_version(installed_version):
                msl_pkgs_to_update[name] = {
                    'installed_version': installed_version,
                    'using_pypi': using_pypi,
                    'extras_require': extras_require,
                    'version': version,
                    'repo_name': repo_name,
                }
            else:
                utils.log.warning('The {!r} package is already the latest [{}]'.format(name, installed_version))
                continue

        w = [max(w[0], len(name+extras_require)), max(w[1], len(installed_version))]

    msl_pkgs_to_update = utils._sort_packages(msl_pkgs_to_update)

    if not msl_pkgs_to_update and not pkgs_non_msl:
        utils.log.info('{0}No packages to update{0}'.format(Fore.RESET))
        return

    msg = ''
    if msl_pkgs_to_update:
        msg += '\n{}The following MSL packages will be {}UPDATED{}:\n'.format(Fore.RESET, Fore.CYAN, Fore.RESET)
        for pkg, info in msl_pkgs_to_update.items():
            pkg += info['extras_require'] + '  '
            msg += '\n  ' + pkg.ljust(w[0]+2) + info['installed_version'].ljust(w[1]) + \
                   ' --> ' + info['version'].replace('==', '') + \
                   '  [{}]'.format('PyPI' if info['using_pypi'] else 'GitHub')

    if pkgs_non_msl:
        if msg:
            msg += '\n'
        msg += '\n{}The following non-MSL packages will be {}UPDATED{}:\n'.format(Fore.RESET, Fore.CYAN, Fore.RESET)
        for pkg, info in pkgs_non_msl.items():
            msg += '\n    ' + pkg.ljust(w_non_msl[0]+2) + info['installed_version'].ljust(w_non_msl[1]) + \
                   ' --> ' + info['version'] + '  [PyPI]'

    utils.log.info(msg)
    if not (yes or utils._ask_proceed()):
        return

    utils.log.info('')

    # If updating the msl-package-manager then update it last
    updating_msl_package_manager = _PKG_NAME in msl_pkgs_to_update
    if updating_msl_package_manager:
        value = msl_pkgs_to_update.pop(_PKG_NAME)
        msl_pkgs_to_update[_PKG_NAME] = value  # using an OrderedDict so this item will be last

    zip_extn = 'zip' if utils._IS_WINDOWS else 'tar.gz'
    exe = [sys.executable, '-m', 'pip', 'install']

    if '--upgrade' not in pip_options or '-U' not in pip_options:
        pip_options.append('--upgrade')
    if '--quiet' not in pip_options or '-q' not in pip_options:
        pip_options.extend(['--quiet'] * utils._pip_quiet)
    if '--disable-pip-version-check' not in pip_options:
        pip_options.append('--disable-pip-version-check')

    # install MSL packages
    for pkg, info in msl_pkgs_to_update.items():
        if info['using_pypi']:
            utils.log.debug('Updating {!r} from PyPI'.format(pkg))
            if info['version'] and info['version'][0] not in '<!=>~':
                info['version'] = '==' + info['version']
            package = [pkg + info['extras_require'] + info['version']]
            pip_github_options = []
        else:
            utils.log.debug('Updating {!r} from GitHub[{}]'.format(pkg, github_suffix))
            if commit or utils.has_git:
                repo = 'git+https://github.com/MSLNZ/{}.git@{}'.format(info['repo_name'], github_suffix)
            else:
                repo = 'https://github.com/MSLNZ/{}/archive/{}.{}'.format(info['repo_name'], github_suffix, zip_extn)
            repo += '#egg={}'.format(pkg)
            pip_github_options = ['--force-reinstall']
            if info['extras_require']:
                repo += info['extras_require']
            else:
                pip_github_options.append('--no-deps')
            package = [repo]

        if utils._IS_WINDOWS and pkg == _PKG_NAME:
            # On Windows, an executable cannot replace itself while it is running. However,
            # an executable can be renamed while it is running. Therefore, we rename msl.exe
            # to msl.exe.old and then a new msl.exe file can be created during the update
            filename = sys.exec_prefix + '/Scripts/msl.exe'
            os.rename(filename, filename + '.old')

        subprocess.call(exe + pip_options + pip_github_options + package)

    # install non-MSL packages
    if pkgs_non_msl:
        packages = []
        for k, v in pkgs_non_msl.items():
            version = v['version']
            if version[0] not in '<!=>~':
                version = ''
            packages.append('{}{}'.format(k, version))
        utils.log.debug('Updating non-MSL packages from PyPI')
        p = subprocess.Popen(exe + pip_options + packages, stderr=subprocess.PIPE)
        _, err = p.communicate()
        if err:
            message = err.decode().rstrip()
            utils.log.error(message)
            pattern = r'requires (\S+), but you have'
            for requires in re.findall(pattern, message):
                utils.log.warning('Rolling back to {!r}'.format(requires))
                subprocess.call(exe + pip_options + [requires])

    if updating_msl_package_manager:
        return 'updating_msl_package_manager'



"""
Command line interface for the :ref:`create <create-cli>` command.
"""
from .utils import log
from .create import create
from .cli_argparse import (
    add_argument_package_names,
    add_argument_yes,
    add_argument_quiet,
    add_argument_disable_mslpm_version_check,
)

HELP = 'Create a new package.'

DESCRIPTION = HELP + """

To create a new package you must specify the package name,
and optionally, the name and email address of the author, the
directory where you to want to save the package and the namespace
of the package.
"""

EXAMPLE = """
Example:
  To create a new package that is part of the msl namespace:
  
  $ msl create MyPackage

  To import the package, use:

  >>> from msl import MyPackage

  To create a package that is part of another namespace, use:

  $ msl create monochromator --namespace pr

  To import this package, use:

  >>> from pr import monochromator
  
  To create a package that is not part of a namespace, use:
  
  $ msl create mypackage --no-namespace
  
  To import this package, use:

  >>> import mypackage

"""


def add_parser_create(parser):
    """Add the :ref:`create <create-cli>` command to the parser."""
    p = parser.add_parser(
        'create',
        help=HELP,
        description=DESCRIPTION,
        epilog=EXAMPLE,
    )
    add_argument_package_names(p)
    add_argument_yes(p)
    p.add_argument(
        '-a', '--author',
        nargs='+',
        help='The name of the author to use for the new package.\n'
             'Can include spaces, for example: --author First Middle Last'
    )
    p.add_argument(
        '-e', '--email',
        help='The email address of the author to use for the new package.'
    )
    p.add_argument(
        '-d', '--dir',
        help='The directory to create the new package(s) in.\n'
             'If not specified then the package(s) will be created\n'
             'in the current working directory.'
    )
    p.add_argument(
        '-n', '--namespace',
        help='The namespace that the package belongs to. Default is "msl".'
    )
    p.add_argument(
        '--no-namespace',
        action='store_true',
        default=False,
        help='Do not create a namespace package.'
    )
    add_argument_quiet(p)
    add_argument_disable_mslpm_version_check(p)
    p.set_defaults(func=execute)


def execute(args, parser):
    """Executes the :ref:`create <create-cli>` command."""
    if args.names:
        if args.no_namespace:
            namespace = None
        elif args.namespace:
            namespace = args.namespace
        else:
            namespace = 'MSL'
        create(*args.names, yes=args.yes, author=args.author, email=args.email,
               directory=args.dir, namespace=namespace)
    else:
        log.error('You must specify the name of the new package')



"""
Install MSL packages.
"""
import sys
import subprocess

from . import utils

# Fixes issue #8 (repository name != package name)
# Not sure how to generalize a universal solution since one is free to choose
# any repository name and package name, therefore this mapping will need to be
# updated on a case-by-case basis in future releases of msl-package-manager
#
# key: repo name, value: package (egg) name
_egg_name_map = {
    'pr-omega-logger': 'omega-logger',
    'pr-single-photons': 'photons',
    'pr-superk-mono': 'superk-mono',
    'pr-webpage-text': 'webpage-text',
    'rpi-ocr': 'ocr',
    'rpi-smartgadget': 'smartgadget',
}


def install(*names, **kwargs):
    """Install MSL packages.

    MSL packages can be installed from PyPI packages_ (only if a release has been
    uploaded to PyPI) or from GitHub repositories_.

    .. note::
       If the MSL packages_ are available on PyPI then PyPI is used as the default
       location to install the package. If you want to force the installation to occur
       from the ``main`` branch from GitHub (even though the package is available on PyPI)
       then set ``branch='main'``. If the package is not available on PyPI
       then the ``main`` branch is used as the default installation location.

    .. _repositories: https://github.com/MSLNZ
    .. _packages: https://pypi.org/search/?q=%22Measurement+Standards+Laboratory+of+New+Zealand%22

    .. versionchanged:: 2.4.0
        Added the `pip_options` keyword argument.

    .. versionchanged:: 2.5.0
        Added the `commit` keyword argument. The default name of a
        repository branch changed to ``main``.

    Parameters
    ----------
    *names
        The name(s) of the MSL package(s) to install. If not specified then
        install all MSL packages that begin with the ``msl-`` prefix. The
        ``msl-`` prefix can be omitted (e.g., ``'loadlib'`` is equivalent to
        ``'msl-loadlib'``). Also accepts shell-style wildcards (e.g., ``'pr-*'``).
    **kwargs
        * branch -- :class:`str`
            The name of a git branch to install. If not specified and neither a
            `tag` nor `commit` was specified then the ``main`` branch is used to
            install a package if it is not available on PyPI.
        * commit -- :class:`str`
            The hash value of a git commit to use to install a package.
        * tag -- :class:`str`
            The name of a git tag to use to install a package.
        * update_cache -- :class:`bool`
            The information about the MSL packages_ that are available on PyPI and about
            the repositories_ that are available on GitHub are cached to use for subsequent
            calls to this function. After 24 hours the cache is automatically updated. Set
            `update_cache` to be :data:`True` to force the cache to be updated when you call
            this function. Default is :data:`False`.
        * yes -- :class:`bool`
            If :data:`True` then don't ask for confirmation before installing.
            The default is :data:`False` (ask before installing).
        * pip_options -- :class:`list` of :class:`str`
            Optional arguments to pass to the ``pip install`` command,
            e.g., ``['--retries', '10', '--user']``

    """
    # TODO Python 2.7 does not support named arguments after using *args
    #  we can define yes=False, branch=None, ...
    #  in the function signature when we choose to drop support for Python 2.7
    utils._check_kwargs(kwargs, {'yes', 'branch', 'commit', 'tag', 'update_cache', 'pip_options'})

    yes = kwargs.get('yes', False)
    branch = kwargs.get('branch', None)
    commit = kwargs.get('commit', None)
    tag = kwargs.get('tag', None)
    update_cache = kwargs.get('update_cache', False)
    pip_options = kwargs.get('pip_options', [])

    if commit and not utils.has_git:
        utils.log.error('Cannot install from a commit because git is not installed')
        return

    github_suffix = utils._get_github_url_suffix(branch=branch, commit=commit, tag=tag)
    if github_suffix is None:
        return

    # keep the order of the log messages consistent: pypi -> github -> local
    # utils._create_install_list() does github -> local
    pkgs_pypi = utils.pypi(update_cache)
    packages = utils._create_install_list(names, branch, commit, tag, update_cache)
    if not packages:
        utils.log.info('No MSL packages to install')
        return

    utils._log_install_uninstall_message(
        packages, 'INSTALLED', branch=branch, commit=commit, tag=tag, pkgs_pypi=pkgs_pypi
    )
    if not (yes or utils._ask_proceed()):
        return

    utils.log.info('')

    zip_extn = 'zip' if utils._IS_WINDOWS else 'tar.gz'
    exe = [sys.executable, '-m', 'pip', 'install']

    if '--quiet' not in pip_options or '-q' not in pip_options:
        pip_options.extend(['--quiet'] * utils._pip_quiet)
    if '--disable-pip-version-check' not in pip_options:
        pip_options.append('--disable-pip-version-check')

    for name, values in packages.items():
        if name in pkgs_pypi and not (branch or commit or tag):
            utils.log.debug('Installing {!r} from PyPI'.format(name))
            if values['extras_require']:
                name += values['extras_require']
            if values['version_requested']:
                name += values['version_requested']
            subprocess.call(exe + pip_options + [name])
        else:
            utils.log.debug('Installing {!r} from GitHub[{}]'.format(name, github_suffix))
            if commit or utils.has_git:
                repo = 'git+https://github.com/MSLNZ/{}.git@{}'.format(name, github_suffix)
            else:
                repo = 'https://github.com/MSLNZ/{}/archive/{}.{}'.format(name, github_suffix, zip_extn)

            egg_name = _egg_name_map.get(name, name)
            repo += '#egg={}'.format(egg_name)
            if values['extras_require']:
                repo += values['extras_require']
            subprocess.call(exe + pip_options + [repo])



"""
Command line interface for the :ref:`update <update-cli>` command.
"""
from .update import update
from .cli_argparse import (
    add_argument_all,
    add_argument_branch,
    add_argument_package_names,
    add_argument_tag,
    add_argument_update_cache,
    add_argument_yes,
    add_argument_quiet,
    add_argument_disable_mslpm_version_check,
    add_argument_commit,
)

DESCRIPTION = """{} MSL packages.

MSL packages are retrieved from GitHub repositories or from PyPI.
"""

EXAMPLE = """All other optional arguments are passed to "pip install --upgrade".

Examples:
    msl {0} equipment qt
    msl {0} loadlib --tag v0.3.0
    msl {0} io --no-deps
"""


def add_parser_update(parser, name='update'):
    """Add the :ref:`update <update-cli>` command to the parser."""
    if name == 'update':
        p = parser.add_parser(
            name,
            help='Update MSL packages.',
            description=DESCRIPTION.format(name.capitalize()),
            epilog=EXAMPLE.format(name),
        )
    else:
        p = parser.add_parser(
            name,
            help='Alias for update.',
            description=DESCRIPTION.format(name.capitalize()),
            epilog=EXAMPLE.format(name),
        )
    add_argument_package_names(p)
    add_argument_all(p)
    add_argument_yes(p)
    add_argument_branch(p)
    add_argument_commit(p)
    add_argument_tag(p)
    add_argument_quiet(p)
    add_argument_update_cache(p)
    add_argument_disable_mslpm_version_check(p)
    p.add_argument(
        '-o', '--non-msl',
        action='store_true',
        default=False,
        help='{} all non-MSL packages that are outdated.\n'
             'Warning, use this option with caution.'.format(name.title())
    )
    p.set_defaults(func=execute)


def execute(args, parser):
    """Executes the :ref:`update <update-cli>` command."""
    if parser.contains_package_names(quiet=args.non_msl) or args.non_msl:
        return update(
            *args.names,
            yes=args.yes,
            branch=args.branch,
            commit=args.commit,
            tag=args.tag,
            update_cache=args.update_cache,
            pip_options=args.pip_options,
            include_non_msl=args.non_msl,
            all_msl=args.all
        )



"""
Install, uninstall, update, list and create MSL packages.

The following constants are available.
"""
import re
from collections import namedtuple

__author__ = 'Measurement Standards Laboratory of New Zealand'
__copyright__ = '\xa9 2017 - 2021, ' + __author__
__version__ = '2.5.2'

_v = re.search(r'(\d+)\.(\d+)\.(\d+)[.-]?(.*)', __version__).groups()

version_info = namedtuple('version_info', 'major minor micro releaselevel')(int(_v[0]), int(_v[1]), int(_v[2]), 'final')
""":obj:`~collections.namedtuple`: Contains the version information as a (major, minor, micro, releaselevel) tuple."""

_PKG_NAME = 'msl-package-manager'

from .utils import (
    pypi,
    github,
    installed,
    info,
    set_log_level,
    outdated_pypi_packages,
)
from .update import update
from .create import create
from .install import install
from .uninstall import uninstall
from .authorise import authorise



import os
import re
import sys
from subprocess import check_output
from setuptools import (
    setup,
    find_packages,
    Command,
)


class ApiDocs(Command):
    """
    A custom command that calls sphinx-apidoc
    see: https://www.sphinx-doc.org/en/master/man/sphinx-apidoc.html
    """
    description = 'builds the api documentation using sphinx-apidoc'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        command = [
            None,  # in Sphinx < 1.7.0 the first command-line argument was parsed, in 1.7.0 it became argv[1:]
            '--force',  # overwrite existing files
            '--module-first',  # put module documentation before submodule documentation
            '--separate',  # put documentation for each module on its own page
            '-o', './docs/_autosummary',  # where to save the output files
            'msl',  # the path to the Python package to document
        ]

        import sphinx
        if sphinx.version_info[:2] < (1, 7):
            from sphinx.apidoc import main
        else:
            from sphinx.ext.apidoc import main
            command.pop(0)

        main(command)


class BuildDocs(Command):
    """
    A custom command that calls sphinx-build
    see: https://www.sphinx-doc.org/en/master/man/sphinx-build.html
    """
    description = 'builds the documentation using sphinx-build'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        command = [
            None,  # in Sphinx < 1.7.0 the first command-line argument was parsed, in 1.7.0 it became argv[1:]
            '-b', 'html',  # the builder to use, e.g., create a HTML version of the documentation
            '-a',  # generate output for all files
            '-E',  # ignore cached files, forces to re-read all source files from disk
            'docs',  # the source directory where the documentation files are located
            './docs/_build/html',  # where to save the output files
        ]

        import sphinx
        if sphinx.version_info[:2] < (1, 7):
            from sphinx import build_main
        else:
            from sphinx.cmd.build import build_main
            command.pop(0)

        build_main(command)


def read(filename):
    with open(filename) as fp:
        return fp.read()


def fetch_init(key):
    # open the __init__.py file to determine a value instead of importing the package
    return re.search(r'{}\s*=\s*(.+)'.format(key), read(init_original)).group(1).strip('\'\"')


def get_version():
    init_version = fetch_init('__version__')
    if 'dev' not in init_version or testing:
        return init_version

    if 'develop' in sys.argv:
        # then installing in editable (develop) mode
        #   python setup.py develop
        #   pip install -e .
        # following PEP-440, the local version identifier starts with '+'
        return init_version + '+editable'

    # append the commit hash to __version__
    setup_dir = os.path.dirname(os.path.realpath(__file__))
    try:
        # write all error messages from git to devnull
        with open(os.devnull, mode='w') as devnull:
            out = check_output(['git', 'rev-parse', 'HEAD'], cwd=setup_dir, stderr=devnull)
            sha1 = out.strip().decode()
    except:
        # the git executable is not available, manually parse .git directory
        try:
            git_dir = os.path.join(setup_dir, '.git')
            with open(os.path.join(git_dir, 'HEAD'), mode='rt') as fp1:
                line = fp1.readline().strip()
                if line.startswith('ref:'):
                    _, ref_path = line.split()
                    with open(os.path.join(git_dir, ref_path), mode='rt') as fp2:
                        sha1 = fp2.readline().strip()
                else:  # detached HEAD
                    sha1 = line
        except:
            return init_version

    suffix = sha1[:7]
    if not suffix or init_version.endswith(suffix):
        return init_version

    # following PEP-440, the local version identifier starts with '+'
    dev_version = init_version + '+' + suffix

    with open(init_original) as fp:
        init_source = fp.read()

    if os.path.isfile(init_backup):
        os.remove(init_backup)
    os.rename(init_original, init_backup)

    with open(init_original, mode='wt') as fp:
        fp.write(re.sub(
            r'__version__\s*=.+',
            "__version__ = '{}'".format(dev_version),
            init_source
        ))

    return dev_version


install_requires = ['setuptools', 'colorama']
tests_require = [
    'pytest>=4.4',  # >=4.4 to support the "-p conftest" option
    'pytest-cov',
]
docs_require = ['sphinx', 'sphinx_rtd_theme']

testing = {'test', 'tests'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if testing else []

needs_sphinx = {'doc', 'docs', 'apidoc', 'apidocs'}.intersection(sys.argv)
sphinx = docs_require + install_requires if needs_sphinx else []

init_original = 'msl/package_manager/__init__.py'
init_backup = init_original + '.backup'
version = get_version()

setup(
    name='msl-package-manager',
    version=version,
    author=fetch_init('__author__'),
    author_email='info@measurement.govt.nz',
    url='https://github.com/MSLNZ/msl-package-manager',
    description='Install, uninstall, update, list and create MSL packages',
    long_description=read('README.rst'),
    platforms='any',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
    ],
    setup_requires=sphinx + pytest_runner,
    tests_require=tests_require,
    install_requires=install_requires,
    extras_require={'tests': tests_require, 'docs': docs_require},
    cmdclass={'docs': BuildDocs, 'apidocs': ApiDocs},
    entry_points={
        'console_scripts': [
            'msl = msl.package_manager.cli:main',
        ],
    },
    packages=find_packages(include=('msl*',)),
    include_package_data=True,
    zip_safe=False,
)

if os.path.isfile(init_backup):
    os.remove(init_original)
    os.rename(init_backup, init_original)
